# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
#
# Copyright © 2014 Stephen Baumgartner <steve@slbaumgartner.com>
#
# Permission to use, copy, modify, and distribute this software for
# any purpose and without fee is hereby granted, provided the above
# copyright notice appears in all copies.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#-----------------------------------------------------------------------------
#
# Name: circle_intersect
#
# Description:
# Calculate the intersection point of two coplanar circles defined by
# three distinct Point3d objects, presumably picked by the user via a
# SketchUp Tool wrapped around this utility:
# 1. a point on the first circle, P1
# 2. the center point of the second circle, C2=P2.  This point is assumed
# to be located on the first circle, so D1 = P2-P1 is a diameter vector of the
# first circle, and the center of the first circle is at
# C1 = P1 + D1/2.
# 3. a point on the second circle, P3.
#
# The radius of the first circle is r1 = |P2 - P1|/2
# The radius of the second circle defaults to r2 = |P3 - P2|, but can also
# be passed as an argument to the initializer.
#
# Provided they are not colinear, these three points define the plane
# containing the circles and their intersection(s), in which 2D formulas
# can be used to find the solution.  If the points are colinear, the
# geometry could be at any rotation about the line of centers and we don't
# have enough info to choose one.
#
# Degenerate cases that are trapped and cause ArgumentException:
# - Points are colinear.  As noted, the solution is ill-defined in this case
# - If P2==P1 or P3==P2 a circle collapses to a point and there is no useful solution
# - If r2==2*r1 the first circle is inset in the second and the intersection
# is a single point at P1, again not useful.  One particular case is if P3==P1.
# - If r2 > 2*r1, there are no real solutions

# The intended "good" case:
# If 0 < r2 < 2*r1 there are two real solutions, symmetric about the line
# of centers between C1 and C2.
#
# Version:       12/21/2014  0.0 initial test of concept
#

require "sketchup.rb"

module SLBPlugins
  module SLBDiagBeam
    class CircleIntersect

      def initialize(p1, p2, p3, radius=nil)
        begin
          set_points(p1,p2,p3, radius)
        rescue ArgumentError => e
          # catch the exceptions from set_points and re-raise
          # them so the caller's code is shown as the location
          # of the error, not here
          raise ArgumentError, e.message, caller[1..-1]
        end
      end

      # separated this from the initializer so that an instance can be
      # reused for multiple solutions if desired
      def set_points(p1,p2,p3, radius=nil)
        @radius = radius
        # Validate that the args are Point3d instances,
        # and that p2 != p1 and p3 != p2 and p3 != p1 else
        # the calculations below may fail
        if p1.class != Geom::Point3d || p2.class != Geom::Point3d || p3.class != Geom::Point3d
          raise ArgumentError, "Arguments must be Point3d", caller
        end
        if(p1 == p2 || p2 == p3 || p3 == p1)
          raise ArgumentError, "Points must be distinct", caller
        end

        # instance vars are clones so that the caller can't tamper
        # with the saved points after initialization
        @p1=p1.clone
        @p2=p2.clone
        @p3=p3.clone

        # radius and diameter unit vector of first circle
        # This unit vector is the "x" direction of the 2D geometry
        dv1=(@p2-@p1)
        @r1=dv1.length/2
        @uv1=dv1.normalize
        # center point of first circle
        offset = Geom::Vector3d.new(@uv1)
        offset.length = @r1
        @cp1= @p1.offset(offset)

        # radius and unit vector of the second circle
        rv2 = @p3-@p2
        # argument to initializer overrides rv2 length
        if @radius.nil?
          @r2 = rv2.length
        else
          @r2=@radius
        end
        uv2 = rv2.normalize
        # center of second circle is p2
        @cp2 = @p2

        # range checks on the circle radii per comments at top of file
        if(@r2 >= 2*@r1)
          raise ArgumentError, "Radius is too large or\n third point is too far from second", caller
        end
        
        # normal vector to the plane of the circles
        # This calculation will produce a zero-length vector if the three points are colinear
        # which means that any orientation of the plane could contain the circles.
        # Rather than choose a plane arbitrarily, trap this case and raise an exception
        @nv = @uv1 * uv2
        if @nv.length == 0
          raise ArgumentError, "Points must not be colinear", caller[1..-1]
        end
        @nv.normalize!
        # unit vector in the plane perpendicular to uv1
        # this vector is the "y" direction of the 2D geometry
        @uv2 = @nv * @uv1
        @uv2.normalize!

        # puts "set_points calculated:"
        # puts "first circle center = #{@cp1.to_a}"
        # puts "first circle radius = #{@r1}"
        # puts "second circle center = #{@cp2.to_a}"
        # puts "second circle radius = #{@r2}"
        # puts "normal vector = #{@nv.to_a}"
        # puts "uv1 = #{@uv1.to_a}"
        # puts "uv2 = #{@uv2.to_a}"

        # nothing returned (i.e. don't leak @uv2)
        nil
      end # set_points

      # find the solution using the specified points.
      # return is an Array of two Point3d objects.
      # The first is the intersection on the same side of the
      # line of centers as p3 and the second is the
      # intersection on the other side.
      def calculate
        # The formulas below could blow up if the tests in set_points weren't
        # made.
        
        # calculate the "x" distance from @cp1 to a line joining the intersection
        # points, per 2D geometry circle formulas.
        x = (2 * @r1**2 - @r2**2)/(2*@r1)

        # and the point on the line of centers where the joining line crosses
        x_offset = Geom::Vector3d.new(@uv1)
        x_offset.length = x
        @xp = @cp1.offset(x_offset)

        # calculate the "y" distance from the line of centers to the intersection
        # points, again per 2D geometry (pythagorean formula)
        y = Math.sqrt(@r1**2 - x**2)

        # and finally, the intersection points
        # ip1 is on the same side of the line of centers as p3
        # ip2 is on the opposite side
        y_offset = Geom::Vector3d.new(@uv2)
        y_offset.length=y
        ip1 = @xp + y_offset
        ip2 = @xp - y_offset

        # return the two solutions in an Array
        [ip1, ip2]
      end

      # attribute accessors.  These return copies so that caller
      # can not possibly mess with our internal state (which would
      # bypass the integrity checks)
      def normal_vector
        return @nv.clone
      end
      
      def first_radius
        @r1
      end

      def first_center
        @cp1.clone
      end

      def second_radius
        @r2
      end

      def second_center
        @cp2.clone
      end

    end # CircleIntersect
  end #SLBDiagBeam
end # SLBPlugins
