# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
#
# Copyright © 2014 Stephen Baumgartner <steve@slbaumgartner.com>
#
# Permission to use, copy, modify, and distribute this software for
# any purpose and without fee is hereby granted, provided the above
# copyright notice appears in all copies.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#-----------------------------------------------------------------------------
#
# Name:           diagonal_beam
# Description:    Take four input points from the user and draw a
#                 rectangular parallelepipid beam from them.
#                 The first two points define the diagonally opposite
#                 corners of the Face of the beam.
#                 The third point defines the plane of the Face.
#                 If the user has not input a width via the VCB,
#                 the distance between the second and third points
#                 is used as the width of the Face.
#                 The fourth point defines the depth/thickness of the
#                 beam normal to the Face.
#                 
# Version:       12/21/2014  0.0 initial test of concept
#                12/29/2014  0.1 rework for better user interaction, more robust
#                                error handling, and overall code structure.
#                1/1/2015    0.2 rework again for better cooperation between draw
#                                and draw_beam.  Also better handling of user input.
#
# Ideas used in this code and help with development were provided by John B (@driven)
#
#-----------------------------------------------------------------------------
require 'sketchup.rb'
require "SB_diagonal_beam/circle_intersect.rb"

module SLBPlugins
  module SLBDiagBeam

    class DiagonalBeam
      DISTINCT_POINTS = "Input points must be distinct" unless defined? DISTINCT_POINTS
      NON_COLINEAR = "Third point must not be on the same line as the first two" unless defined? NON_COLINEAR
      NON_ZERO_THICKNESS = "Fourth point must not be on the face plane" unless defined? NON_ZERO_THICKNESS
      IS_MAC = (RUBY_PLATFORM =~ /darwin/) unless defined? IS_MAC
      RETAIN_MSG = IS_MAC ? "press option to unstick" : "press tab to unstick" unless defined? RETAIN_MSG
      NON_RETAIN_MSG = IS_MAC ? "press option to stick" : "press tab to stick" unless defined? NON_RETAIN_MSG

      def initialize
        # @ip is the InputPoint from the current pick
        @ip = Sketchup::InputPoint.new
        # @ip1 is the InputPoint from the previous pick
        # this is used as reference when @ip picks, which, in principle, should
        # cause the inference engine to favor points on the object containing the
        # previous pick.  Not clear that this technique works!
        @ip1 = Sketchup::InputPoint.new

        @drawn = false

        # By default, previous values are retained until the tool is deactivated.
        # When @retain is false, they are cleared when the user starts a new
        # beam by picking the first point.  They can also be cleared by entering
        # blank in the VCB
        @retain = true

        # Instance of SLBPlugins::SLBDiagBeam::CircleIntersect that will be reused for geometry
        # calculations.  Don't have enough info here to create one, but initialize
        # the attribute to nil as a trap in case code below tries to use it before one
        # is properly initialized.
        @ci=nil
        
        # user can input sizes by clicking or via VCB.  Need to keep straight
        # where the current values came from
        @user_width = nil
        @beam_width = nil
        @user_depth = nil
        @beam_depth = nil

        # @pts[0] = first picked corner point of face
        # @pts[1] = diagonal opposite corner point of face
        # @pts[2] = point picked to set face rectangle width
        # @pts[3] = point picked to set beam depth, projected onto normal line of face
        @pts = []

        # @state = 0 => waiting for first corner pick
        # @state = 1 => waiting for diagonal corner pick
        # @state = 2 => waiting for width pick
        # @state = 3 => waiting for depth pick
        @state = 0

        @cursor = nil   # make sure @cursor exists regardless of whether cursor_path is found
        if(Sketchup.version.to_i >= 16)
          if IS_MAC
            cursor_path=Sketchup.find_support_file("diag_beam.pdf", "Plugins/SB_diagonal_beam")
          else
            cursor_path=Sketchup.find_support_file("diag_beam.svg", "Plugins/SB_diagonal_beam")
          end
        else
          cursor_path = Sketchup.find_support_file("diag_beam_16.png", "Plugins/diagonal_beam")
        end
        # for those who like giant cursors :)
        #      cursor_path = Sketchup.find_support_file("diag_beam_32.png", "Plugins/diagonal_beam")
        
        if(cursor_path)
          @cursor =  UI::create_cursor(cursor_path, 0, 20)
          #        @cursor =  UI::create_cursor(cursor_path, 0, 31)
        end
        UI.set_cursor(@cursor)
      end # initialize

      # clear out saved data
      # by default, prior beam width and depth are not reset
      # if purge_dims is true, they are also cleared to nil.
      def reset(purge_dims = false)
        @pts = []
        @state = 0
        @beam_width = nil if purge_dims
        @beam_depth = nil if purge_dims
        @ip1.clear
        @drawn = false
        UI.set_cursor(@cursor)
        show_status
      end #reset

      # capture and validate the user's pick point
      # an error message box is posted if there is a problem
      # the return is true if ok and false if bad input
      def capture_input_point
        @ip1.copy! @ip
        pick_point = @ip.position
        
        # validate the pick point
        valid = false
        case @state
        when 0
          # no constraints on the first point
          valid = true
        when 1
          # the second point must not be the same as the first
          valid = (pick_point != @pts[0])
          msg = DISTINCT_POINTS
        when 2
          # third point must not be on same line as first two
          # else we can't calculate a plane or width.  Test
          # these conditions by trying to create the normal vector
          v1 = pick_point - @pts[1]
          v2 = @pts[1] - @pts[0]
          @face_normal = v1*v2
          valid = @face_normal.length != 0

          # capture the user's selected width
          if(valid)
            if(! @user_width.nil?)
              @beam_width = @user_width
            else
              if(@retain && !@beam_width.nil?)
              # retain the prior value
              else
                # either not retaining or no prior value - capture from the pick
                @beam_width = v1.length
              end
            end
          end
          msg = NON_COLINEAR
        when 3
          # fourth point must not be in the plane of the face
          # else we have no thickness
          pick_point = pick_point.project_to_line(@normal_line)
          face_plane = [@pts[0], @face_normal]
          depth = pick_point.distance_to_plane(face_plane)
          valid = (depth != 0)

          # capture the user's selected depth
          if(valid)
            if(! @user_depth.nil?)
              @beam_depth = @user_depth
            else
              if(@retain && !@beam_depth.nil?)
              else
                @beam_depth = depth
              end
            end
            msg = NON_ZERO_THICKNESS
          end
        end # case @state
        
        if(valid)
          # ok to save the input point
          @pts[@state] = pick_point
        else
          # tell the user what's wrong
          UI.messagebox(msg)
        end
        return valid
      end # capture_input_point

      # perform actions associated with going from @state to @state+1
      def process_state_transition
        case @state
        when 0
          # User has picked the first point.  Starting a new beam, so clear out prior values
          @pts[1] = @pts[2] = @pts[3] = nil
          @drawn = false
          unless @retain
            # clear out prior width and depth when @retain is false
            @user_width = nil
            @beam_width=nil
            @user_depth = nil
            @beam_depth=nil
          end
        when 1
        # we have the two corners of the face diagonal.  Enough info for draw to show
        # the diagonal, but not enough to know the face plane or width yet
        when 2
          # we have three points - enough to define the initial face.
          # create @ci for reuse everywhere if not already done
          if(@ci.nil?)
            # This should be the first and only time we need to create a new CircleIntersect
            # All other situations should be able to reuse @ci via set_points.
            @ci=SLBPlugins::SLBDiagBeam::CircleIntersect.new(@pts[0], @pts[1], @pts[2], @beam_width)
          else
            # when we have a pre-existing @ci, make sure it has the right data for calculate
            @ci.set_points(@pts[0], @pts[1], @pts[2], @beam_width)
          end
          # create a line through @pts[1] normal to the face to use while calculating depth
          @normal_line = [@pts[1], @face_normal]
        when 3
          # have all we need - draw the beam
          draw_beam
        end
        increment_state
      end

      # advance to the next state and set the status texts appropriately.
      def increment_state
        if(@state < 3)
          @state += 1
        else
          @state = 0
        end
        show_status
      end

      # draw the beam
      # All the necessary point locations have already been created during the draw method,
      # so here we need to create equivalent SketchUp Entities inside an undoable operation.
      def draw_beam

        # start the drawing operation
        model = Sketchup.active_model
        model.start_operation("Diagonal Beam")
        
        # work in the currently active context so the new geometry will be inside
        # a Group or ComponentInstance if one is open instead of the global context
        ents = model.active_entities

        # make a new group to hold the beam
        grp = ents.add_group

        # draw a Face in the group and make sure it is oriented correctly for pushpull
        face = grp.entities.add_face(@pts[0], @corner_1, @pts[1], @corner_2)
        face.reverse! unless face.normal.samedirection? @depth_offset

        # pushpull the Face to the required depth
        face.pushpull(@depth_offset.length)

        # TBD: explode the group so that the new Face interacts with pre-existing
        # geometry?  The user can do so manually, but if that will be more common
        # than leaving the group, we could do it here.

        # tell undo the end of the bundled operation
        model.commit_operation

        @drawn=true
      end # draw_beam

      ############################################
      # event handlers called by SketchUp Tool protocol
      
      def activate
        reset
      end

      def deactivate(view)
        reset
        # make sure we don't leave behind any helper lines on the view
        view.invalidate
        view.refresh
      end

      # make VCB accept typed input without needing click in it first
      # Note: prior to SU 2015 this method did not seem to be needed -
      # the default was true.
      def enableVCB?
        return true
      end

      # need this to cause inference point and tooltips to be drawn as the mouse moves
      def onMouseMove(flags, x, y, view)
        @ip.pick(view, x, y, @ip1)
        view.invalidate
        view.refresh
      end

      # make sure our cursor is visible
      def onSetCursor
        UI.set_cursor(@cursor)
      end
      
      # user clicks the mouse button - capture the data point and advance the state
      def onLButtonDown(flags, x, y, view)
        @ip.pick(view, x, y, @ip1)
        if(capture_input_point)
          process_state_transition
          view.refresh
        end
      end

      # user cancelled this Tool
      # according to the API docs:
      # flag = 0 -> hit ESC
      # flag = 1 -> reselected same Tool (this does not appear to ever happen!
      #             instead we get a deactivate followed by an activate
      # flag = 2 -> undo while Tool active
      def onCancel(flag, view)
        view.invalidate
        view.refresh
        reset(false)
      end

      # toggle value of @retain when user presses ALT/option on Mac, Tab on Windows
      def onKeyDown(key, rpt, flags, view)
        if(RUBY_PLATFORM =~ /darwin/)
          keycode = VK_ALT
        else
          keycode = 9
        end
        if key == keycode
          @retain = !@retain
        end
        # refresh status to indicate the value of @retain
        show_status
      end #onKeyDown
      
      # Accept user input in the VCB as the desired width and depth of the beam
      # The expected format is two values with a separator between.  To allow for
      # varying locale conventions on format of decimals, the separator may be one of
      # "x","*",";",or ",".
      # Leading and/or trailing whitespace is allowed on both values
      # To input only a width, the separator and depth may be omitted
      # To input only a depth (retaining current value of width) a blank
      # followed by a separator and depth value may be entered.
      # To clear both width and depth, enter a blank.
      def onUserText(text, view)

        # blank means reset to calculating width from points regardless of
        # value of @retain
        if(text == "")
          @user_width=nil
          @beam_width=nil
          @user_depth=nil
          @beam_depth=nil
          return
        end

        input = []
        value1=nil
        value2=nil
        
        begin
          # accept a generous mix of input separators to allow for locales
          if text =~ /x/
            input = text.split('x')
          elsif text =~ /\*/
            input = text.split('*')
          elsif text =~ /\;/
            input = text.split(';')
          else text =~ /\,/ && !text =~ /\;/ or !text =~ /\./
            input = text.split(',')
          end
          
          # remove leading and trailing whitespace
          input[0].strip! unless input[0].nil?
          input[1].strip! unless input[1].nil?

          # try to convert both fields to lengths to trap bad inputs
          value1 = input[0].to_l unless input[0].nil?
          value2 = input[1].to_l unless input[1].nil?
        rescue
          # Error parsing the text into valid lengths
          UI.messagebox("please enter valid number(s) or blank to reset")
          Sketchup::set_status_text "", SB_VCB_VALUE
          return
        end

        # succeeded parsing inputs as lengths.  Now see if values are acceptable
        # always check the width input
        if(@beam_width.nil? && (value1.nil? || value1 <= 0.0))
          UI.messagebox("width must be a positive number")
          Sketchup::set_status_text "", SB_VCB_VALUE
          return
        end
        # check the depth input only when present 
        if(input.length == 2 && @beam_depth.nil? && (value2.nil? || value2 <= 0.0))
          UI.messagebox("depth must be a positive number")
          Sketchup::set_status_text "", SB_VCB_VALUE
          return
        end
        
        if input.length == 1
          # only width was input
          @user_width = value1
          @beam_width = @user_width
          # this will redraw our helper lines if appropriate
          view.invalidate
          view.refresh
        elsif input.length == 2
          # either both fields input, or width skipped as in "x5"
          if(value1 != 0)
            # width was present
            @user_width = value1
            @beam_width = @user_width
          end
          if(value2 != 0)
            @user_depth = value2
            @beam_depth = @user_depth
          end
          # this will redraw our helper lines if appropriate
          view.invalidate
          view.refresh
        end
        show_status
        
        # redraw at new radius, if appropriate
        if @drawn
          # undo the previous operation before redrawing
          # this avoids buildup of obsolete operations on the
          # undo stack
          Sketchup.undo
          # update the calculated locations of things
          find_corners
          find_depth_offset
          draw_beam
        end
      end # onUserText

      ##################################
      # screen updaters

      # set the status line and VCB prompts
      # arguments override current values of state so that draw
      # can dynamically track width and depth without implicitly
      # modifying @beam_width or @beam_depth
      def show_status(width_arg=nil, depth_arg=nil)

        status_width = (width_arg.nil?) ? @beam_width : width_arg
        status_width = status_width.to_l unless status_width.nil?
        status_depth = (depth_arg.nil?) ? @beam_depth : depth_arg
        status_depth = status_depth.to_l unless status_depth.nil?

        # VCB willing to accept and show width and depth in any state
        Sketchup::set_status_text "Width x Depth", SB_VCB_LABEL
        lead = @retain? "[" : ""
        trail = @retain? "]" : ""

        # depth might be blank or nil, but depth without width not allowed?
        if(!status_width.nil? && status_width != 0)
          Sketchup::set_status_text "#{lead}#{status_width} x #{status_depth}#{trail}", SB_VCB_VALUE
        else
          Sketchup::set_status_text "", SB_VCB_VALUE
        end

        # state-dependent prompts
        if(@retain)
          msg = RETAIN_MSG
        else
          msg = NON_RETAIN_MSG
        end
        case @state
        when 0
          Sketchup::set_status_text "Select first corner (#{msg})", SB_PROMPT
        when 1
          Sketchup::set_status_text "Select diagonal corner (#{msg})", SB_PROMPT
        when 2
          Sketchup::set_status_text "Select a point (for face plane and for width) (#{msg})", SB_PROMPT
        when 3
          Sketchup::set_status_text "Select a point (for depth) (#{msg})", SB_PROMPT
        end

      end

      # invoked by SketchUp when the view is redrawn.
      # draws our dynamic hints lines.  These are 2D OpenGL
      # objects, not SketchUp entities.  Make sure to invalidate
      # the view when this Tool exits, else they will persist!
      def draw(view)
        # make sure the inference engine's tooltip is visible
        view.tooltip = @ip.tooltip
        @ip.draw view

        case @state
        when 0
          show_status
        when 1
          # draw "rubber band" to cursor from first point
          view.drawing_color = "magenta"
          view.line_width = 2
          view.draw(GL_LINE_STRIP,  @pts[0], @ip.position)
          show_status
        when 2
          # draw the diagonal
          view.drawing_color = "cyan"
          view.line_width = 2
          view.draw(GL_LINE_STRIP, @pts[0], @pts[1])
          
          # draw "rubber band" to cursor from third point
          view.drawing_color = "magenta"
          view.line_width = 2
          view.draw(GL_LINE_STRIP,  @pts[1], @ip.position)

          # show the current width in the VCB unless value locked
          if @retain && !@beam_width.nil?
            show_status
          else
            show_status((@ip.position - @pts[1]).length)
          end
        when 3
          @pts[3] = @ip.position.project_to_line(@normal_line)
          find_corners
          find_depth_offset
          
          # draw face of beam
          view.drawing_color = "magenta"
          view.line_width = 2
          view.draw GL_LINE_LOOP, [@corner_2, @pts[0], @corner_1, @pts[1]]

          # draw "rubber band" skeleton of the beam body
          if(@depth_offset.length > 0)
            gl1 = @pts[1].offset(@depth_offset)
            gl2 = @pts[0].offset(@depth_offset)
            gl3 = @corner_1.offset(@depth_offset)
            gl4 = @corner_2.offset(@depth_offset)
            # all projected lines
            view.draw(GL_LINE_STRIP,  @pts[1], gl1)
            view.draw(GL_LINE_STRIP, @pts[0], gl2)
            view.draw(GL_LINE_STRIP, @corner_1, gl3)
            view.draw(GL_LINE_STRIP, @corner_2, gl4)
            # second loop
            view.draw GL_LINE_LOOP, [gl4, gl2, gl3, gl1]

            # show the current width and depth in the VCB
            if @retain && !@beam_depth.nil?
              show_status
            else
              show_status(nil, @depth_offset.length) 
            end
          end
        end # case @state
      end

      # Use CircleIntersect to find the locations of the other two corners of the beam face
      # This returns values by setting @corner_1 and @corner_2 for use when drawing or creating entities
      # It assumes that @pts[0], @pts[1], and @pts[2] have all been captured previously,
      # that @beam_width has been calculated, and that @ci exists
      def find_corners
        width = @beam_width
        begin
          # need to keep @ci up to date with @beam_width in case user has entered a new one in VCB
          @ci.set_points(@pts[0], @pts[1], @pts[2], @beam_width)
          intersections = @ci.calculate
        rescue ArgumentError => e
          # @ci may throw ArgumentError if the points are duplicates, are co-linear or if width is too
          # large to give a solution.  capture_input_point already tested for dups and co-linear,
          # so we need to take steps here to recover if width is too big.  This is
          # purely a problem with CircleIntersect's algorithm, so we can recover by
          # chopping down the width until it allows a solution.  Note: the potential exists for
          # multiple retries until width is acceptable, which will cause sluggish behavior of
          # draw.  However, repeated decimating of width guarantees that eventually it will be
          # small enough!
          width=width/10
          retry
        end
        @corner_1 = intersections[0]
        @corner_2 = @pts[0] + (@pts[1] - @corner_1)
      end # find_corners

      # calculate beam depth offset vector based on last picked point
      # This assumes that @pts[1] and @pts[3] are meaningful
      # It adjusts the offset vector length based on state
      # Return is implicitly via @depth_offset for use when drawing or creating entities
      def find_depth_offset
        offset = @pts[3] - @pts[1]

        if(offset.length != 0)
          # have something to draw only if depth isn't zero
          if(!@user_depth.nil?)
            offset.length = @user_depth
          else
            if(@retain && !@beam_depth.nil?)
              offset.length = @beam_depth
            else
              # no action needed - using pick depth as is
            end
          end
        end
        @depth_offset = offset
      end # find_depth_offset

    end # class DiagonalBeam

    def self.diagonal_beam_tool
      Sketchup.active_model.select_tool SLBPlugins::SLBDiagBeam::DiagonalBeam.new
    end

  end # SLBDiagBeam

end # SLBPlugins

#=============================================================================

if( not file_loaded?("diagonal_beam.rb") )
  cmd = UI::Command.new("Diagonal Beam") {SLBPlugins::SLBDiagBeam::diagonal_beam_tool}
  cmd.menu_text = "Diagonal Beam"
  if(Sketchup.version.to_i >= 16)
    if(RUBY_PLATFORM =~ /darwin/)
      cmd.small_icon = Sketchup.find_support_file("diag_beam.pdf", "Plugins/SB_diagonal_beam")
      cmd.large_icon = Sketchup.find_support_file("diag_beam.pdf", "Plugins/SB_diagonal_beam")
    else
      cmd.small_icon = Sketchup.find_support_file("diag_beam.svg", "Plugins/SB_diagonal_beam")
      cmd.large_icon = Sketchup.find_support_file("diag_beam.svg", "Plugins/SB_diagonal_beam")
    end
  else
    cmd.small_icon = Sketchup.find_support_file("diag_beam_16.png", "Plugins/SB_diagonal_beam")
    cmd.large_icon = Sketchup.find_support_file("diag_beam_24.png", "Plugins/SB_diagonal_beam")
  end
  UI.menu("Tools").add_item cmd

  toolbar = UI::Toolbar.new("Diagonal Beam")
  toolbar.add_item(cmd)
  toolbar.restore if toolbar.get_last_state==1


end
#-----------------------------------------------------------------------------
file_loaded("diagonal_beam.rb")


